/* DASubprocess.m created by DAspinall on Thu 27-Aug-1998 */

#import "DASubprocess.h"
#import "NSString+Comp.h"

NSString *LaunchPath			= @"LaunchPath";
NSString *CurrentDirectoryPath	= @"CurrentDirectoryPath";
NSString *Arguments				= @"Arguments";
NSString *StandardError			= @"StandardError";
NSString *StandardOutput		= @"StandardOutput";
NSString *TerminationStatus		= @"TerminationStatus";

@implementation DASubprocess

/* Class method to execute a command or series of pipe separated commands.  Feeds the standard output from each command into the standard input of the next command.  Stops on the first command that fails to execute correctly
	Example: "ls -l |grep Application | wc -c"
*/
+ (NSDictionary *)launchProcesses:(NSString *)commands
{
    NSMutableDictionary *results = [NSMutableDictionary dictionary];
    NSCharacterSet *pipeSet = [NSCharacterSet characterSetWithCharactersInString:@"|"];
    NSArray *cmdLines = [commands componentsSeparatedBySet:pipeSet];

    if ( cmdLines && ([cmdLines count] > 0))
    {
        NSDictionary *previousRslts = nil;
        NSEnumerator *cmdEnum = [cmdLines objectEnumerator];
        NSString *line = nil;
        BOOL errors = NO;

        while ((errors == NO) && (line = [cmdEnum nextObject]))
        {
            DASubprocess *process = [[[DASubprocess alloc] init] autorelease];
            NSMutableArray *cmdArgs = [[[line componentsSeparatedBySet:[NSCharacterSet whitespaceCharacterSet]] mutableCopy] autorelease];
            NSString *input = [previousRslts objectForKey:StandardOutput];

            [process setLaunchPath:[cmdArgs objectAtIndex:0]];
            [cmdArgs removeObjectAtIndex:0];
            [process setArguments:cmdArgs];

            NS_DURING

            [process launchWithInput:input];
            [process waitUntilExit];

            NS_HANDLER
            NS_ENDHANDLER

            previousRslts = [process resultDictionary];

            if ( [[previousRslts objectForKey:TerminationStatus] intValue] != 0)
                errors = YES;
        }

        [results addEntriesFromDictionary:previousRslts];
        [results setObject:commands forKey:LaunchPath];
    }
    return results;
}

- init
{
    [super init];
    launched = NO;
    outAvailable = NO;
    errAvailable = NO;
    task = [[NSTask alloc] init];
    [self setEnvironment:[[NSProcessInfo processInfo] environment]];
    return self;
}

- (void)dealloc
{
    [task release];
    [stdOutData release];
    [stdErrData release];
    [exceptionMessage release];
    [super dealloc];
}

- (void)launch
{
    [self launchWithInput:nil];
}

/* Launch the task, and optionally pass the string 'input' to the subtask as standard input.  Input is optional and may be an empty string or nil.  Catches NSTask exceptions and sets the TerminationStatus to -1
NOTE: this is the only method that exception wraps.  Any exception raised by the underlying NSTask in the wrap methods will not be caught. */
- (void)launchWithInput:(NSString *)input
{
    if ( launched == NO )
    {
        NSPipe			*outPipe = nil;
        NSFileHandle	*outHandle = nil;

        NSPipe			*errPipe = nil;
        NSFileHandle	*errHandle = nil;

        NSPipe			*inPipe = nil;
        NSFileHandle	*inHandle = nil;

        outPipe = [NSPipe pipe];
        outHandle = [outPipe fileHandleForReading];
        [task setStandardOutput:outPipe];

        errPipe = [NSPipe pipe];
        errHandle = [errPipe fileHandleForReading];
        [task setStandardError:errPipe];

        inPipe = [NSPipe pipe];
        inHandle = [inPipe fileHandleForWriting];
        [task setStandardInput:inPipe];

        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(readStdOut:) name:NSFileHandleReadToEndOfFileCompletionNotification object:outHandle];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(readStdErr:) name:NSFileHandleReadToEndOfFileCompletionNotification object:errHandle];

        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(taskCompleted:) name:NSTaskDidTerminateNotification object:task];

        {
            NSArray *modes = [NSArray arrayWithObjects:NSDefaultRunLoopMode, @"DASubprocess_RunLoopMode", nil];
            [outHandle readToEndOfFileInBackgroundAndNotifyForModes:modes];
            [errHandle readToEndOfFileInBackgroundAndNotifyForModes:modes];
        }

        NS_DURING
            
        outAvailable = YES;
        errAvailable = YES;
        launched = YES;
        [task launch];

        NS_HANDLER
        {
            outAvailable = NO;
            errAvailable = NO;

            exceptionMessage = [[NSString alloc] initWithFormat:@"Caught exception: %@ (%@).\n\"%@\"", [localException name], [localException reason], [task launchPath]];

            NSLog( @"[%s %s] %@",  __FILE__,  sel_getName(_cmd), exceptionMessage);
        }
        NS_ENDHANDLER
        
        if (input != nil && ([input length] > 0))
        {
            [inHandle writeData:[input dataUsingEncoding:[NSString defaultCStringEncoding] allowLossyConversion:YES]];
        }
        [inHandle closeFile];

        
        // one loop will usually complete the task
        [[NSRunLoop currentRunLoop] runMode:@"DASubprocess_RunLoopMode" beforeDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];
    }
    else
        [NSException raise:NSGenericException format:@"%@ Cannot relaunch task", NSStringFromClass([self class])];
}

// here we provide 2 controlled runloops.  The first "DASubprocess_RunLoopMode" alows the NSFileHandles to continue reading, the second allows NSNotifications to be sent.  We flip between them until both pipes are closed and the task is completed.  Note, the task can complete before the pipes have closed, so wit for both events.
- (void)waitUntilExit
{
    while ((outAvailable == YES) || (errAvailable == YES) || ([task isRunning] == YES))
    {
        // give the NSFileHandles (pipes) to finish reading
        [[NSRunLoop currentRunLoop] runMode:@"DASubprocess_RunLoopMode" beforeDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];

        // Give the task a chance to exit and notify
        [[NSRunLoop currentRunLoop] runMode:NSDefaultRunLoopMode beforeDate:[NSDate dateWithTimeIntervalSinceNow:0.5]];
    }
}

- (void)taskCompleted:(NSNotification *)notice
{
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSTaskDidTerminateNotification object:task];
}

- (void)readStdOut:(NSNotification *)notice
{
    stdOutData = [[notice userInfo] objectForKey:NSFileHandleNotificationDataItem];
    [stdOutData retain];
    outAvailable = NO;
    
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSFileHandleReadCompletionNotification object:[notice object]];
}

- (void)readStdErr:(NSNotification *)notice
{
    stdErrData = [[notice userInfo] objectForKey:NSFileHandleNotificationDataItem];
    [stdErrData retain];
    errAvailable = NO;

    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSFileHandleReadCompletionNotification object:[notice object]];
}

- (NSDictionary *)resultDictionary
{
    NSMutableDictionary *dictionary = [NSMutableDictionary dictionary];

    if ( launched == NO )
    {
        [dictionary setObject:@"Process not launched" forKey:StandardError];
        [dictionary setObject:[NSNumber numberWithInt:(-1)] forKey:TerminationStatus];
    }
    else if ([task isRunning] == YES)
    {
        [dictionary setObject:@"Process still running" forKey:StandardError];
        [dictionary setObject:[NSNumber numberWithInt:(-1)] forKey:TerminationStatus];
    }
    else
    {
        NSString *string = nil;
        [dictionary setObject:[task launchPath] forKey:LaunchPath];
       
        if ([task currentDirectoryPath])
            [dictionary setObject:[task currentDirectoryPath] forKey:CurrentDirectoryPath];
        
        if ([task arguments])
            [dictionary setObject:[task arguments] forKey:Arguments];

        string = [[[NSString alloc] initWithData:stdErrData encoding:NSASCIIStringEncoding] autorelease];

        if ( exceptionMessage )
        {
            [dictionary setObject:[NSString stringWithFormat:@"%@\n%@", exceptionMessage, string] forKey:StandardError];
        }
        else
            [dictionary setObject:string forKey:StandardError];
        
        [dictionary setObject:[[[NSString alloc] initWithData:stdOutData encoding:NSASCIIStringEncoding] autorelease] forKey:StandardOutput];

        NS_DURING
            [dictionary setObject:[NSNumber numberWithInt:[task terminationStatus]] forKey:TerminationStatus];
        NS_HANDLER
            [dictionary setObject:[NSNumber numberWithInt:(-1)] forKey:TerminationStatus];
        NS_ENDHANDLER
    }

    return dictionary;
}

// get the raw data
- (NSData *)stdOutData
{
    return stdOutData;
}

- (NSData *)stdErrData
{
    return stdErrData;
}

// search for the command from the task environment.  Stop as soon as you find an executable.
- (NSString *)findPathForCommand:(NSString *)command
{
    NSFileManager *fileManager = [NSFileManager defaultManager];
    NSDictionary *env = [task environment];
    NSString *fullPath = command;
    
    if ( env )
    {
        NSString *pathEnv = [env objectForKey:@"PATH"];
        if (pathEnv == nil)
            pathEnv = [env objectForKey:@"Path"];

        if (pathEnv != nil)
        {
            NSArray			*pathArray = nil;
            NSEnumerator	*pathEnum = nil;
            NSString		*aPath = nil;
#ifdef WIN32
            pathArray = [pathEnv componentsSeparatedByString:@";"];
#else
            pathArray = [pathEnv componentsSeparatedByString:@":"];
#endif
            pathEnum = [pathArray objectEnumerator];
            
            while (aPath = [pathEnum nextObject])
            {
                NSString		*tempPath = nil;
                
                tempPath = [aPath stringByAppendingPathComponent:command];
                if ([fileManager isExecutableFileAtPath:tempPath] == YES)
                {
                    fullPath = tempPath;
                    break;
                }
#ifdef WIN32
                else if ([[tempPath pathExtension] isEqualToString:@"exe"] == NO)
                {
                    tempPath = [tempPath stringByAppendingPathExtension:@"exe"];
                    if ([fileManager isExecutableFileAtPath:tempPath] == YES)
                    {
                        fullPath = tempPath;
                        break;
                    }
                }
#endif
            }
        }
    }

    return fullPath;
}
// NSTask cover methods
- (void)setLaunchPath:(NSString *)path
{
    if ( launched == NO )
    {
        NSString *exePath = path;
    #ifdef WIN32
        if ([[exePath pathExtension] isEqualToString:@"exe"] == NO)
            exePath = [path stringByAppendingPathExtension:@"exe"];
    #endif

        if ([exePath isAbsolutePath] == NO)
        {
            if ([[NSFileManager defaultManager] isExecutableFileAtPath:exePath] == NO)
            {
                exePath = [self findPathForCommand:path];
            }
        }
        [task setLaunchPath:exePath];
    }
    else
        [NSException raise:NSGenericException format:@"%@ Cannot change task launch path after task has been launched", NSStringFromClass([self class])];
}

- (void)setArguments:(NSArray *)arguments
{
    if ( launched == NO )
        [task setArguments:(NSArray *)arguments];
    else
        [NSException raise:NSGenericException format:@"%@ Cannot change task arguments after task has been launched", NSStringFromClass([self class])];
}


- (void)setEnvironment:(NSDictionary *)dict
{
    if ( launched == NO )
        [task setEnvironment:(NSDictionary *)dict];
    else
        [NSException raise:NSGenericException format:@"%@ Cannot change task environment after task has been launched", NSStringFromClass([self class])];
}


- (void)setCurrentDirectoryPath:(NSString *)path
{
    if ( launched == NO )
        [task setCurrentDirectoryPath:(NSString *)path];
    else
        [NSException raise:NSGenericException format:@"%@ Cannot change task current working directory after task has been launched", NSStringFromClass([self class])];
}

- (BOOL)isRunning
{
    return [task isRunning];
}

- (NSString *)launchPath
{
    return [task launchPath];
}

- (NSArray *)arguments
{
    return [task arguments];
}

- (NSDictionary *)environment
{
    return [task environment];
}

- (NSString *)currentDirectoryPath
{
    return [task currentDirectoryPath];
}

- (int)terminationStatus
{
    return [task terminationStatus];
}

- (void)interrupt
{
    [task interrupt];
}

- (void)terminate
{
    [task terminate];
}

@end
